@@ -6,19 +6,7 @@ import tempfile
 import pytest
 
 import pyvips
-from helpers import \
-    IMAGES, JPEG_FILE, SRGB_FILE, MATLAB_FILE, PNG_FILE, TIF_FILE, OME_FILE, \
-    ANALYZE_FILE, GIF_FILE, WEBP_FILE, EXR_FILE, FITS_FILE, OPENSLIDE_FILE, \
-    PDF_FILE, SVG_FILE, SVGZ_FILE, SVG_GZ_FILE, GIF_ANIM_FILE, DICOM_FILE, \
-    BMP_FILE, NIFTI_FILE, ICO_FILE, CUR_FILE, TGA_FILE, SGI_FILE, AVIF_FILE, \
-    AVIF_FILE_HUGE, TRUNCATED_FILE, \
-    GIF_ANIM_EXPECTED_PNG_FILE, GIF_ANIM_DISPOSE_BACKGROUND_FILE, \
-    GIF_ANIM_DISPOSE_BACKGROUND_EXPECTED_PNG_FILE, \
-    GIF_ANIM_DISPOSE_PREVIOUS_FILE, \
-    GIF_ANIM_DISPOSE_PREVIOUS_EXPECTED_PNG_FILE, \
-    temp_filename, assert_almost_equal_objects, have, skip_if_no, \
-    TIF1_FILE, TIF2_FILE, TIF4_FILE, WEBP_LOOKS_LIKE_SVG_FILE, \
-    WEBP_ANIMATED_FILE, JP2K_FILE, RGBA_FILE
+from helpers import *
 
 class TestForeign:
     tempdir = None
@@ -32,8 +20,6 @@ class TestForeign:
         cls.mono = cls.colour.extract_band(1).copy()
         # we remove the ICC profile: the RGB one will no longer be appropriate
         cls.mono.remove("icc-profile-data")
-        cls.rad = cls.colour.float2rad().copy()
-        cls.rad.remove("icc-profile-data")
         cls.cmyk = cls.colour.bandjoin(cls.mono)
         cls.cmyk = cls.cmyk.copy(interpretation=pyvips.Interpretation.CMYK)
         cls.cmyk.remove("icc-profile-data")
@@ -41,15 +27,23 @@ class TestForeign:
         im = pyvips.Image.new_from_file(GIF_FILE)
         cls.onebit = im[1] > 128
 
+        all = [cls.mono, cls.colour, cls.cmyk]
+        # and alpha variants of all of them
+        alpha = [x.bandjoin(255) for x in all]
+        # and with a second alpha
+        alpha2 = [x.bandjoin(255) for x in alpha]
+
+        cls.all = all + alpha + alpha2
+
     @classmethod
     def teardown_class(cls):
         shutil.rmtree(cls.tempdir, ignore_errors=True)
         cls.colour = None
         cls.rgba = None
         cls.mono = None
-        cls.rad = None
         cls.cmyk = None
         cls.onebit = None
+        cls.all = None
 
     # we have test files for formats which have a clear standard
     def file_loader(self, loader, test_file, validate):
@@ -163,6 +157,10 @@ class TestForeign:
         self.save_load_buffer("jpegsave_buffer", "jpegload_buffer",
                               self.colour, 80)
 
+        for image in self.all:
+            target = pyvips.Target.new_to_memory()
+            image.jpegsave_target(target)
+
         # see if we have exif parsing: our test image has this field
         x = pyvips.Image.new_from_file(JPEG_FILE)
         if x.get_typeof("exif-ifd0-Orientation") != 0:
@@ -427,6 +425,7 @@ class TestForeign:
             assert im.height == 442
             assert im.bands == 3
             assert im.get("bits-per-sample") == 16
+            assert im.get_typeof("palette") == 0
 
         self.file_loader("pngload", PNG_FILE, png_valid)
         self.buffer_loader("pngload_buffer", PNG_FILE, png_valid)
@@ -436,6 +435,23 @@ class TestForeign:
         self.save_load_file(".png", "[interlace]", self.colour)
         self.save_load_file(".png", "[interlace]", self.mono)
 
+        for image in self.all:
+            target = pyvips.Target.new_to_memory()
+            image.pngsave_target(target)
+
+        def png_indexed_valid(im):
+            a = im(10, 10)
+            assert_almost_equal_objects(a, [148.0, 131.0, 109.0])
+            assert im.width == 290
+            assert im.height == 442
+            assert im.bands == 3
+            assert im.get("bits-per-sample") == 8
+            assert im.get("palette") == 1
+
+        self.file_loader("pngload", PNG_INDEXED_FILE, png_indexed_valid)
+        self.buffer_loader("pngload_buffer",
+            PNG_INDEXED_FILE, png_indexed_valid)
+
         # size of a regular mono PNG
         len_mono = len(self.mono.write_to_buffer(".png"))
 
@@ -484,7 +500,32 @@ class TestForeign:
             "exif-ifd0-ImageDescription", "test description")
         im2 = pyvips.Image.new_from_buffer(
             im1.write_to_buffer(".png"), "")
-        assert im2.get("exif-ifd0-ImageDescription").startswith("test description")
+        assert im2.get("exif-ifd0-ImageDescription") \
+            .startswith("test description")
+
+        # https://github.com/libvips/libvips/issues/4509
+        data = self.rgba.cast("double").write_to_buffer(".png")
+        after = pyvips.Image.new_from_buffer(data, "")
+        assert (self.rgba - after).abs().max() == 0
+
+        # we should be able to save an 8-bit image as a 16-bit PNG
+        data = self.colour.pngsave_buffer(bitdepth=16)
+        rgb16 = pyvips.Image.pngload_buffer(data)
+        assert rgb16.format == "ushort"
+
+        # we should be able to save a 16-bit image as an 8-bit PNG
+        data = rgb16.pngsave_buffer(bitdepth=8)
+        rgb = pyvips.Image.pngload_buffer(data)
+        assert rgb.format == "uchar"
+
+        # we should be able to save a 16-bit image as an 8-bit WebP
+        if have("webpsave"):
+            data = rgb16.webpsave_buffer(lossless=True)
+            rgb = pyvips.Image.webpload_buffer(data)
+            assert rgb.format == "uchar"
+            # ... and check if it was correctly shifted down
+            # https://github.com/libvips/libvips/issues/4568
+            assert (self.colour - rgb).abs().max() == 0
 
     @skip_if_no("tiffload")
     def test_tiff(self):
@@ -499,6 +540,10 @@ class TestForeign:
         self.file_loader("tiffload", TIF_FILE, tiff_valid)
         self.buffer_loader("tiffload_buffer", TIF_FILE, tiff_valid)
 
+        for image in self.all:
+            target = pyvips.Target.new_to_memory()
+            image.tiffsave_target(target)
+
         def tiff1_valid(im):
             a = im(127, 0)
             assert_almost_equal_objects(a, [0.0])
@@ -535,6 +580,17 @@ class TestForeign:
 
         self.file_loader("tiffload", TIF4_FILE, tiff4_valid)
 
+        def tiff_subsampled_valid(im):
+            a = im(10, 10)
+            assert_almost_equal_objects(a, [6.0, 5.0, 21.0, 255.0])
+            assert im.width == 250
+            assert im.height == 325
+            assert im.bands == 4
+            assert im.get("bits-per-sample") == 8
+
+        self.file_loader("tiffload", TIF_SUBSAMPLED_FILE, tiff_subsampled_valid)
+        self.buffer_loader("tiffload_buffer", TIF_SUBSAMPLED_FILE, tiff_subsampled_valid)
+
         self.save_load_buffer("tiffsave_buffer", "tiffload_buffer", self.colour)
         self.save_load("%s.tif", self.mono)
         self.save_load("%s.tif", self.colour)
@@ -631,6 +687,15 @@ class TestForeign:
         assert x1.xres == 100
         assert x1.yres == 200
 
+        if sys.platform == "darwin":
+            with open(TIF2_FILE, 'rb') as f:
+                buf = bytearray(f.read())
+            buf = buf[:-4]
+            source = pyvips.Source.new_from_memory(buf)
+            im = pyvips.Image.tiffload_source(source, fail_on="warning")
+            with pytest.raises(Exception) as e_info:
+                im.avg() > 0
+
         # OME support in 8.5
         x = pyvips.Image.new_from_file(OME_FILE)
         assert x.width == 439
@@ -694,6 +759,31 @@ class TestForeign:
             z = y.hist_find(band=0)
             assert z(0, 0)[0] + z(255, 0)[0] == y.width * y.height
 
+    @skip_if_no("tiffload")
+    @pytest.mark.xfail(raises=AssertionError, reason="fails when libtiff was configured with --disable-old-jpeg")
+    def test_tiff_ojpeg(self):
+        def tiff_ojpeg_tile_valid(im):
+            a = im(10, 10)
+            assert_almost_equal_objects(a, [135.0, 156.0, 177.0, 255.0])
+            assert im.width == 234
+            assert im.height == 213
+            assert im.bands == 4
+            assert im.get("bits-per-sample") == 8
+
+        self.file_loader("tiffload", TIF_OJPEG_TILE_FILE, tiff_ojpeg_tile_valid)
+        self.buffer_loader("tiffload_buffer", TIF_OJPEG_TILE_FILE, tiff_ojpeg_tile_valid)
+
+        def tiff_ojpeg_strip_valid(im):
+            a = im(10, 10)
+            assert_almost_equal_objects(a, [228.0, 15.0, 9.0, 255.0])
+            assert im.width == 160
+            assert im.height == 160
+            assert im.bands == 4
+            assert im.get("bits-per-sample") == 8
+
+        self.file_loader("tiffload", TIF_OJPEG_STRIP_FILE, tiff_ojpeg_strip_valid)
+        self.buffer_loader("tiffload_buffer", TIF_OJPEG_STRIP_FILE, tiff_ojpeg_strip_valid)
+
     @skip_if_no("jp2kload")
     @skip_if_no("tiffload")
     def test_tiffjp2k(self):
@@ -740,16 +830,14 @@ class TestForeign:
         assert im.height == height * 5
 
         # page/n let you pick a range of pages
-        # 'n' param added in 8.5
-        if pyvips.at_least_libvips(8, 5):
-            im = pyvips.Image.magickload(GIF_ANIM_FILE)
-            width = im.width
-            height = im.height
-            im = pyvips.Image.magickload(GIF_ANIM_FILE, page=1, n=2)
-            assert im.width == width
-            assert im.height == height * 2
-            page_height = im.get("page-height")
-            assert page_height == height
+        im = pyvips.Image.magickload(GIF_ANIM_FILE)
+        width = im.width
+        height = im.height
+        im = pyvips.Image.magickload(GIF_ANIM_FILE, page=1, n=2)
+        assert im.width == width
+        assert im.height == height * 2
+        page_height = im.get("page-height")
+        assert page_height == height
 
         # should work for dicom
         im = pyvips.Image.magickload(DICOM_FILE)
@@ -905,6 +993,11 @@ class TestForeign:
         assert x.height == 16731
         buf = x.webpsave_buffer()
 
+        # target_size should reasonably work, +/- 2% is fine
+        x = pyvips.Image.new_from_file(WEBP_FILE)
+        buf_size = len(x.webpsave_buffer(target_size=20_000, keep='none'))
+        assert 19600 < buf_size < 20400
+
     @skip_if_no("analyzeload")
     def test_analyzeload(self):
         def analyze_valid(im):
@@ -967,6 +1060,7 @@ class TestForeign:
 
     @skip_if_no("openslideload")
     def test_openslideload(self):
+
         def openslide_valid(im):
             a = im(10, 10)
             assert_almost_equal_objects(a, [244, 250, 243, 255])
@@ -1020,13 +1114,17 @@ class TestForeign:
         assert x2.get("background") == [81, 81, 81]
         assert x2.get("interlaced") == 1
         assert x2.get("bits-per-sample") == 4
+        assert x2.get("palette") == 1
 
-        x2 = pyvips.Image.new_from_file(GIF_ANIM_FILE, n=-1)
+        x2 = pyvips.Image.new_from_file(GIF_ANIM_FILE,
+                                        n=-1,
+                                        access="sequential")
         # our test gif has delay 0 for the first frame set in error
         assert x2.get("delay") == [0, 50, 50, 50, 50]
         assert x2.get("loop") == 32761
         assert x2.get("background") == [255, 255, 255]
         assert x2.get_typeof("interlaced") == 0
+        assert x2.get("palette") == 1
         # test deprecated fields too
         assert x2.get("gif-loop") == 32760
         assert x2.get("gif-delay") == 0
@@ -1049,6 +1147,10 @@ class TestForeign:
         x2 = pyvips.Image.new_from_file(GIF_ANIM_FILE, page=1, n=-1)
         assert x2.height == 4 * x1.height
 
+        with pytest.raises(pyvips.error.Error):
+            x1 = pyvips.Image.new_from_file(GIF_ANIM_FILE_INVALID, n=-1)
+            x1.avg()
+
     @skip_if_no("gifload")
     def test_gifload_animation_dispose_background(self):
         x1 = pyvips.Image.new_from_file(GIF_ANIM_DISPOSE_BACKGROUND_FILE, n=-1)
@@ -1147,41 +1249,71 @@ class TestForeign:
         assert im.width == 10
         assert im.height == 10
 
+        # Custom CSS stylesheet
+        im = pyvips.Image.new_from_file(SVG_FILE)
+        assert im.avg() < 5
+        im = pyvips.Image.new_from_file(SVG_FILE, stylesheet=b'path{stroke:#f00;stroke-width:1em;}')
+        assert im.avg() > 5
+
     def test_csv(self):
         self.save_load("%s.csv", self.mono)
 
     def test_matrix(self):
         self.save_load("%s.mat", self.mono)
 
+        for image in self.all:
+            target = pyvips.Target.new_to_memory()
+            image.matrixsave_target(target)
+
     @skip_if_no("ppmload")
     def test_ppm(self):
-        self.save_load("%s.ppm", self.mono)
         self.save_load("%s.ppm", self.colour)
 
-        self.save_load_file("%s.ppm", "[ascii]", self.mono, 0)
+        self.save_load_file("%s.pgm", "[ascii]", self.mono, 0)
         self.save_load_file("%s.ppm", "[ascii]", self.colour, 0)
 
-        self.save_load_file("%s.ppm", "[ascii,bitdepth=1]", self.onebit, 0)
+        self.save_load_file("%s.pbm", "[ascii]", self.onebit, 0)
 
         rgb16 = self.colour.colourspace("rgb16")
         grey16 = self.mono.colourspace("rgb16")
 
-        self.save_load("%s.ppm", grey16)
         self.save_load("%s.ppm", rgb16)
 
         self.save_load_file("%s.ppm", "[ascii]", grey16, 0)
         self.save_load_file("%s.ppm", "[ascii]", rgb16, 0)
 
-        source = pyvips.Source.new_from_memory(b'P1\n#\n#\n1 1\n0\n')
-        im = pyvips.Image.ppmload_source(source)
+        im = pyvips.Image.new_from_buffer(b'P1\n#\n#\n1 1\n0\n', "")
         assert im.height == 1
         assert im.width == 1
 
     @skip_if_no("radload")
     def test_rad(self):
-        self.save_load("%s.hdr", self.colour)
+        def hdr_valid(im):
+            # might still be in RADIANCE coding
+            if im.coding == "rad":
+                im = im.rad2float()
+
+            assert_almost_equal_objects(im(10, 10),
+                                        [0.533, 0.564, 0.693],
+                                        threshold=0.01)
+
+            assert_almost_equal_objects(im(1592, 855),
+                                        [1580, 1364, 1276],
+                                        threshold=0.01)
+
+            assert im.width == 1655
+            assert im.height == 1764
+            assert im.bands == 3
+
+        self.file_loader("radload", RAD_FILE, hdr_valid)
+        self.buffer_loader("radload_buffer", RAD_FILE, hdr_valid)
+
+        rad = pyvips.Image.radload(RAD_FILE)
+        self.save_load("%s.hdr", rad)
+        self.save_load("%s.pfm", rad)
+        self.save_load("%s.tif", rad)
         self.save_buffer_tempfile("radsave_buffer", ".hdr",
-                                  self.rad, max_diff=0)
+                                  rad, max_diff=0)
 
     @skip_if_no("dzsave")
     def test_dzsave(self):
@@ -1275,6 +1407,27 @@ class TestForeign:
         assert x.width == 256
         assert x.height == 256
 
+        # IIIF v2
+        im = pyvips.Image.black(3509, 2506, bands=3)
+        filename = temp_filename(self.tempdir, '')
+        im.dzsave(filename, layout="iiif")
+        assert os.path.exists(filename + "/info.json")
+        assert os.path.exists(filename + "/0,0,512,512/512,/0/default.jpg")
+        assert os.path.exists(filename + "/2560,2048,512,458/512,/0/default.jpg")
+        x = pyvips.Image.new_from_file(filename + "/full/439,/0/default.jpg")
+        assert x.width == 439
+        assert x.height == 314
+
+        # IIIF v3
+        filename = temp_filename(self.tempdir, '')
+        im.dzsave(filename, layout="iiif3")
+        assert os.path.exists(filename + "/info.json")
+        assert os.path.exists(filename + "/0,0,512,512/512,512/0/default.jpg")
+        assert os.path.exists(filename + "/2560,2048,512,458/512,458/0/default.jpg")
+        x = pyvips.Image.new_from_file(filename + "/full/439,314/0/default.jpg")
+        assert x.width == 439
+        assert x.height == 314
+
         # test zip output
         filename = temp_filename(self.tempdir, '.zip')
         self.colour.dzsave(filename)
@@ -1340,7 +1493,7 @@ class TestForeign:
         # test keep=pyvips.ForeignKeep.ICC ... icc profiles should be
         # passed down
         filename = temp_filename(self.tempdir, '')
-        self.colour.dzsave(filename, keep=1 << 3) # pyvips.ForeignKeep.ICC
+        self.colour.dzsave(filename, keep=pyvips.ForeignKeep.ICC)
 
         y = pyvips.Image.new_from_file(filename + "_files/0/0_0.jpeg")
         assert y.get_typeof("icc-profile-data") != 0
@@ -1363,9 +1516,6 @@ class TestForeign:
             im = pyvips.Image.heifload(AVIF_FILE_HUGE)
             assert im.avg() == 0.0
 
-        im = pyvips.Image.heifload(AVIF_FILE_HUGE, unlimited=True)
-        assert im.avg() == 0.0
-
     @skip_if_no("heifsave")
     def test_avifsave(self):
         self.save_load_buffer("heifsave_buffer", "heifload_buffer",
@@ -1377,9 +1527,7 @@ class TestForeign:
         im = pyvips.Image.new_from_file(AVIF_FILE)
         buf = im.heifsave_buffer(effort=0, lossless=True, compression="av1")
         im2 = pyvips.Image.new_from_buffer(buf, "")
-        # requires libheif >= 1.13.0 for true lossless:
-        # see: https://github.com/strukturag/libheif/commit/b2612dd9c63f8835cf2047960b8cacd464a325a4
-        assert (im - im2).abs().max() <= 1.0
+        assert (im - im2).abs().max() == 0
 
     @skip_if_no("heifsave")
     def test_avifsave_Q(self):
@@ -1423,6 +1571,7 @@ class TestForeign:
             assert y.get("exif-ifd0-XPComment").startswith("banana")
 
     @skip_if_no("heifsave")
+    @pytest.mark.xfail(raises=pyvips.error.Error, reason="requires libheif built with patent-encumbered HEVC dependencies")
     def test_heicsave_16_to_12(self):
         rgb16 = self.colour.colourspace("rgb16")
         data = rgb16.heifsave_buffer(lossless=True)
@@ -1436,6 +1585,7 @@ class TestForeign:
         assert((im - rgb16).abs().max() < 4500)
 
     @skip_if_no("heifsave")
+    @pytest.mark.xfail(raises=pyvips.error.Error, reason="requires libheif built with patent-encumbered HEVC dependencies")
     def test_heicsave_16_to_8(self):
         rgb16 = self.colour.colourspace("rgb16")
         data = rgb16.heifsave_buffer(lossless=True, bitdepth=8)
@@ -1449,6 +1599,7 @@ class TestForeign:
         assert((im - rgb16 / 256).abs().max() < 80)
 
     @skip_if_no("heifsave")
+    @pytest.mark.xfail(raises=pyvips.error.Error, reason="requires libheif built with patent-encumbered HEVC dependencies")
     def test_heicsave_8_to_16(self):
         data = self.colour.heifsave_buffer(lossless=True, bitdepth=12)
         im = pyvips.Image.heifload_buffer(data)
@@ -1473,6 +1624,27 @@ class TestForeign:
         self.file_loader("jp2kload", JP2K_FILE, jp2k_valid)
         self.buffer_loader("jp2kload_buffer", JP2K_FILE, jp2k_valid)
 
+        # Bretagne2_4.j2k is a tiled jpeg2000 image with 127x127 pixel tiles,
+        # triggering tricky rounding issues
+        filename = os.path.join(IMAGES, "Bretagne2_4.j2k")
+        im4 = pyvips.Image.new_from_file(filename, page=4)
+        im3 = pyvips.Image.new_from_file(filename, page=3)
+        assert abs(im4.avg() - im3.avg()) < 0.2
+
+        # Bretagne2_1.j2k is an untiled jpeg2000 image with non-zero offset
+        filename = os.path.join(IMAGES, "Bretagne2_1.j2k")
+        im4 = pyvips.Image.new_from_file(filename, page=4)
+        im3 = pyvips.Image.new_from_file(filename, page=3)
+        assert abs(im4.avg() - im3.avg()) < 0.5
+
+        # this horrible thing has a header that doesn't match the decoded
+        # pixels ... although it's a valid jp2k image, we reject files of
+        # this type
+        filename = os.path.join(IMAGES, "issue412.jp2")
+        with pytest.raises(Exception) as e_info:
+            im = pyvips.Image.new_from_file(filename)
+            im.avg()
+
     @skip_if_no("jp2ksave")
     def test_jp2ksave(self):
         self.save_load_buffer("jp2ksave_buffer", "jp2kload_buffer",
@@ -1555,6 +1727,7 @@ class TestForeign:
         lossless = self.colour.jxlsave_buffer(lossless=True)
         assert len(lossy) < len(lossless) / 5
 
+    @skip_if_no("gifload")
     @skip_if_no("gifsave")
     def test_gifsave(self):
         # Animated GIF round trip
